#!/usr/bin/perl -w
#
# createfusebits32u4.pl
#
# Perl script to calculate the lock and fuse bytes for a ATmega32u4
#
# To use set the parameters in this source file as described in the comments
# below. Result is printed to STDOUT.
#
# Version 1.0   2018 Jan 18th   jost.brachert@gmx.de
#
# Copyright (C) 2018  Jost Brachert, jost.brachert@gmx.de
#
# This program is free software; you can redistribute it and/or  modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the license, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program, see file COPYING; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
# Boston, MA  02110-1301, USA.
#

use strict;


my $i;

#==============================================================================
# LOCK BITS
#
# Atmel-7766J-USB-ATmega16U4/32U4-Datasheet_04/2016
#   http://www.atmel.com/devices/atmega32u4.aspx
#   28.1 Program And Data Memory Lock Bits

#------------------------------------------------------------------------------
# Lock Bit Settings
#
# Note: Lock bits can only be erased to 1 by Chip Erase command.
#

# lock 0-1  Lock Bits
#           Set exactly one of the elements to one, all others to zero
my @LB = (
    0,  # Flash & EEPROM programming and verification disabled, boot bits & fuses locked
    0,  # Reserved
    0,  # Flash & EEPROM programming disabled, fuses locked
    1,  # No locks enabled
);

# lock 2-3  Boot Lock Bit Protection Mode 0
#           Set exactly one of the elements to one, all others to zero
my @BLB0 = (
    0,  # No appl programming by SPM instruction (Store Program Memory),
        # no appl read by boot lock LPM
    0,  # No appl read by boot lock LPM instruction (Load Program Memory)
    0,  # No appl programming by SPM
    1,  # No appl programming restrictions
);

# lock 4-5  Boot Lock Bit Protection Mode 1
#           Set exactly one of the elements to one, all others to zero
my @BLB1 = (
    0,  # No boot loader programming by SPM instruction (Store Program Memory),
        # no boot loader read by appl LPM
    0,  # No boot loader read by appl LPM instruction (Load Program Memory)
    1,  # No boot loader programming by SPM
    0,  # No boot loader programming restrictions
);

# Set to 1 if to be programmed:
my $lock6 = 1;      # Reserved
my $lock7 = 1;      # Reserved

#------------------------------------------------------------------------------
# Lock Byte Assembly

    my $locks = 0;

    for ($i=0; $i<@LB; ++$i) {
        last if $LB[$i];
    }
    $locks |= $i;

    for ($i=0; $i<@BLB0; ++$i) {
        last if $BLB0[$i];
    }
    $locks |= $i << 2;

    for ($i=0; $i<@BLB1; ++$i) {
        last if $BLB1[$i];
    }
    $locks |= $i << 4;

    $locks |= ( ! $lock6) << 6;
    $locks |= ( ! $lock7) << 7;


#==============================================================================
# LOW FUSE BYTE
#
# Atmel-7766J-USB-ATmega16U4/32U4-Datasheet_04/2016
#   http://www.atmel.com/devices/atmega32u4.aspx
#   28.2 Fuse Bits

#------------------------------------------------------------------------------
# Low Fuse Byte Settings
#
#   Displayed bit values here: 0->programmed, 1->erased
#
# - - - - - - - - - - - - - - - - - - - - -   SUT CKSEL
# External Clock                                   0000
# Reserved                                         0001
# Internal Calibrated RC Oscillator                0010
# Reserved                                         0011
# Start Up Time (SUT) recommended for
#  BOD enabled                                 00
#  Fast rising power                           01
#  Slowly rising power                         10
#  Reserved                                    11
#
# - - - - - - - - - - - - - - - - - - - - -   SUT CKSEL
# Low-Frequency Crystal Oscillator                 010x
#  Start Up Time 1K CK                                0
#  Start Up Time 32K CK                               1
# Additional Delay (by Watchdog Oscillator)
#  4CK       Fast rising power or BOD enabled 00
#  4CK+4.1ms Slowly rising power               01
#  4CK+65ms  Stable frequency at start-up      10
#  Reserved                                    11
#
# Reserved                                         0110
# Reserved                                         0111
#
# Low Power Crystal Oscillator
#     0.4 -  0.9 MHz                               100x
#     0.9 -  3.0 MHz                               101x
#     3.0 -  8.0 MHz                               110x
#     8.0 - 16.0 MHz                               111x
#
# Low Power Crystal Oscillator:
# Ceramic resonator, fast rising power         00     0
# Ceramic resonator, slowly rising power       01     0
# Ceramic resonator, BOD enabled               10     0
# Ceramic resonator, fast rising power         11     0
# Ceramic resonator, slowly rising power       00     1
# Crystal Oscillator, BOD enabled              01     1
# Crystal Oscillator, fast rising power        10     1
# Crystal Oscillator, slowly rising power      11     1

# !! Set bit to 0 if to be programmed
my $SUT   =   0b11;
my $CKSEL = 0b1111;

# Set to 1 if to be programmed:
my $CKOUT  = 0;     # 1: Enable clock output
my $CKDIV8 = 0;     # 1: Divide clock by 8

#------------------------------------------------------------------------------
# Low Fuse Byte Assembly

    my $lfuse = ($SUT << 4) | $CKSEL;
    $lfuse |= ( ! $CKOUT ) << 6;
    $lfuse |= ( ! $CKDIV8) << 7;


#==============================================================================
# HIGH FUSE BYTE
#
# Atmel-7766J-USB-ATmega16U4/32U4-Datasheet_04/2016
#   http://www.atmel.com/devices/atmega32u4.aspx
#   28.2 Fuse Bits

#------------------------------------------------------------------------------
# High Fuse Byte Settings

my @BOOTSZ = (  # Set exactly one of the elements to one, all others to zero:
    1,  # 2048 words
    0,  # 1024 words
    0,  #  512 words
    0,  #  256 words
);

# Set to 1 if to be programmed:
my $BOOTRST = 1;    # Select Reset Vector Adress
                    # 1: Set to start of boot flash section
my $EESAVE  = 0;    # 1: EEPROM memory is preserved through the Chip Erase
my $WDTON   = 0;    # 1: Watchdog Timer always on
my $SPIEN   = 1;    # 1: Enable Serial Program and Data Downloading
my $JTAGEN  = 0;    # 1: Enable JTAG
my $OCDEN   = 0;    # 1: Enable On Chip Debugging

#------------------------------------------------------------------------------
# High Fuse Byte Assembly

    my $hfuse = 0;

    for ($i=0; $i<@BOOTSZ; ++$i) {
        last if $BOOTSZ[$i];
    }
    $hfuse |= $i << 1;

    $hfuse |= ( ! $BOOTRST) << 0;
    $hfuse |= ( ! $EESAVE ) << 3;
    $hfuse |= ( ! $WDTON  ) << 4;
    $hfuse |= ( ! $SPIEN  ) << 5;
    $hfuse |= ( ! $JTAGEN ) << 6;
    $hfuse |= ( ! $OCDEN  ) << 7;


#==============================================================================
# EXTENDED FUSE BYTE
#
# Atmel-7766J-USB-ATmega16U4/32U4-Datasheet_04/2016
#   http://www.atmel.com/devices/atmega32u4.aspx
#   28.2 Fuse Bits

#------------------------------------------------------------------------------
# Extended Fuse Byte Settings

my @BODLEVEL = (    # Set exactly one of the elements to one, all others to zero:
    0,  # VBOTtyp 4.3V       # 000 (0 here means programmed, 1 erased)
    0,  # VBOTtyp 3.5V       # 001
    0,  # VBOTtyp 3.4V       # 010
    1,  # VBOTtyp 2.6V       # 011
    0,  # VBOTtyp 2.4V       # 100
    0,  # VBOTtyp 2.2V       # 101
    0,  # VBOTtyp 2.0V       # 110
    0,  # BOD disabled       # 111
);

# Set to 1 if to be programmed:
my $HWBE   = 0;     # Hardware Boot Enable
my $efuse4 = 1;     # Reserved
my $efuse5 = 1;     # Reserved
my $efuse6 = 0;     # Reserved
my $efuse7 = 0;     # Reserved

#------------------------------------------------------------------------------
# Extended Fuse Byte Assembly

    my $efuse = 0;

    for ($i=0; $i<@BODLEVEL; ++$i) {
        last if $BODLEVEL[$i];
    }
    $efuse |= $i;

    $efuse |= ( ! $HWBE   ) << 3;
    $efuse |= ( ! $efuse4 ) << 4;
    $efuse |= ( ! $efuse5 ) << 5;
    $efuse |= ( ! $efuse6 ) << 6;
    $efuse |= ( ! $efuse7 ) << 7;


#==============================================================================
# OUTPUT

printf "locks:    %3d    0x%02x    0b%08b\n", $locks, $locks, $locks;
printf "lfuse:    %3d    0x%02x    0b%08b\n", $lfuse, $lfuse, $lfuse;
printf "hfuse:    %3d    0x%02x    0b%08b\n", $hfuse, $hfuse, $hfuse;
printf "efuse:    %3d    0x%02x    0b%08b\n", $efuse, $efuse, $efuse;
